#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sched.h>
#include <map>

#define UNUSED(x)    ((void)x)

typedef struct {
   cpu_set_t initial_cpu_affinity;
   cpu_set_t current_cpu_affinity;
   int current_cpu_count;
   int last_update_cycle;
} ProcessInfo;

typedef std::map<int,ProcessInfo> ProcessMap;
ProcessMap g_pids;
int g_current_cycle = 1;
int g_cpus = 1;

void update_current_pids( void )
{
   g_current_cycle++;

   //open /proc directory
   DIR *dir = opendir("/proc");

   if (!dir)
   {
      fprintf(stderr,"Cannot read directory /proc: ");
      perror("");
      exit(1);
   }

   struct dirent *entry;
   //walk each entry
   while( (entry = readdir(dir)) != NULL )
   {
      //int num_span = strspn(entry->d_name,"0123456789");

      //printf("Handling dir entry (type=%04X, span=%d, strlen=%d)\"%s\"\n",entry->d_type,num_span,strlen(entry->d_name),entry->d_name);
      if (    (entry->d_type == DT_DIR)                                    //directory entry
           && (entry->d_name[strspn(entry->d_name,"0123456789")] == 0)     //only numbers (pid)
         )
      {
         int pid = atoi(entry->d_name);
         ProcessInfo *proc_info = &g_pids[pid];

         if (proc_info->last_update_cycle == 0)
         {
            //printf("Recording new pid: %d\n",pid);
            if (sched_getaffinity(pid,sizeof(cpu_set_t),&proc_info->initial_cpu_affinity) != 0)
            {
               int error_id = errno;
               fprintf(stderr,"Failed getting current affinity (pid=%d): %s\n",pid,strerror(error_id));
            }

            proc_info->current_cpu_affinity = proc_info->initial_cpu_affinity;
         }
         else
         {
            sched_getaffinity(pid,sizeof(cpu_set_t),&proc_info->current_cpu_affinity);
         }

         proc_info->current_cpu_count = 0;
         for(int i=0; i<CPU_SETSIZE; i++)
         {
            if (CPU_ISSET(i,&proc_info->current_cpu_affinity))
            {
               proc_info->current_cpu_count++;
               if (g_cpus < i+1)
               {
                  g_cpus = i+1;
               }
            }
         }

         proc_info->last_update_cycle = g_current_cycle;
      }
   }

   closedir(dir);
}

void show_cpu_bindings( void )
{
   ProcessMap::iterator it = g_pids.begin();

   printf("Per Process CPU Map:\n");
   printf("--------------------\n");
   while(it != g_pids.end())
   {
      printf("PID %6d: ",it->first);
      for(int i=0; i<g_cpus; i++)
      {
         printf("%d",CPU_ISSET(i,&it->second.current_cpu_affinity));
      }
      printf("\n");

      ++it;
   }
   printf("\n");
}

void shield_cpu( int cpu )
{
   ProcessMap::iterator it = g_pids.begin();
   int safe_cpu;

   if (cpu == 0)
      safe_cpu = g_cpus-1;    //use last cpu
   else
      safe_cpu = 0;

   while(it != g_pids.end())
   {
      ProcessInfo *info = &it->second;

      if ( CPU_ISSET(cpu,&info->current_cpu_affinity) )
      {
         CPU_CLR(cpu,&info->current_cpu_affinity);

         if ( info->current_cpu_count == 1 )
         {
            CPU_SET(safe_cpu,&info->current_cpu_affinity);
         }
         else
         {
            info->current_cpu_count--;
         }

         if (sched_setaffinity(it->first,sizeof(cpu_set_t),&info->current_cpu_affinity) != 0)
         {
            int error_id = errno;
            fprintf(stderr,"Failed shielding CPU%d from process %d: %s\n",cpu,it->first,strerror(error_id));
         }

         //try to get the current cpu affinity
         sched_getaffinity(it->first,sizeof(cpu_set_t),&info->current_cpu_affinity);
         info->current_cpu_count = 0;
         for(int i=0; i<CPU_SETSIZE; i++)
         {
            if (CPU_ISSET(i,&info->current_cpu_affinity))
            {
               info->current_cpu_count++;
               if (g_cpus < i+1)
               {
                  g_cpus = i+1;
               }
            }
         }

      }

      ++it;
   }
   printf("\n");
}

void restore_cpu_affinities( void )
{
   ProcessMap::iterator it = g_pids.begin();

   while(it != g_pids.end())
   {
      if ( sched_setaffinity(it->first,sizeof(cpu_set_t),&it->second.initial_cpu_affinity) != 0 )
      {
         int error_id = errno;
         fprintf(stderr,"Failed restoring CPU affinity for process %d: %s\n",it->first,strerror(error_id));
      }
      ++it;
   }
}


int main( int argc, char **argv )
{
   UNUSED(argc);
   UNUSED(argv);

   update_current_pids();

   if (g_cpus <= 1)
   {
      fprintf(stderr,"shielding is only possible on multi-core systems\n");
      return 1;
   }

   //show_cpu_bindings();

   shield_cpu(1);
   //show_cpu_bindings();
   system("cat /proc/interrupts");
   getchar();
   system("cat /proc/interrupts");

   restore_cpu_affinities();

   return 0;
}
